/*****************************************************************************/
/*
                               qdLogStats.c

Quick-and-Dirty Log Statistics.

Utility to extract very elementary statistics from Web server common/combined
format log files.  A number of filters allow subsets of the log contents to be
selected.  These filters are simple sort-of regular expressions, not
case-sensitive (deliberately), can contain wildcards (such as asterisks (*),
question marks (?), and percent signs (%)) as well as sort-of regular
expressions (such as the range [a-z]).  THERE IS NO WAY (that I know of) TO
ESCAPE THESE RESERVED CHARACTERS! This functionality uses decc$match_wild()
function.  When using filters it is necessary to provide wildcard 'match-alls'
at the start and/or end of the string if the searched-for string does not fall
at the start and/or end of the field.  Hence to find all records having a field
containing "fred" anywhere in the string the following must be used

  "*fred*"

Special constructs allow more complex expressions to be built up.  Combinations
of required and excluded strings may be specified in the one expression.  When
a string begins with a "+{" it *must* be present for the record NOT to be
filtered out.  If it begins "-{" then it must *not* be present.  Such
specifications must be terminated with a matching closure "}".  For example,
with the following expression

  "+{*test*}-{*fred*}"

the record field must contain a string "test" somewhere but cannot contain the
string "fred" anywhere.  The expression

  "+{*one*}+{*two*}"

specifies that the strings "one" and "two" must be present in the field but not
necessarily in any particular order.  An expression

  "-{*three*}-{*four*}"

excludes the field if it contains either of the two fields anywhere in it.

Matching log file records can be viewed as processed using the
/VIEW=[type] qualifier, where <type> is ALL, MATCH (default) or NOMATCH.

A progress indicator can be displayed using the /PROGRESS qualifier.  With this
a "+" is output for each file processed and a "." for each 1000 records in any
file.  An optional integer can be provides as /PROGRESS=integer to change this
record value.

Note that all selections are based on string matching.  Files and/or records
cannot be selected on the basis of being before or after a particular date for
instance.  Also that records are accepted or rejected in the order in which the
results are displayed, that is if a record is rejected against method it will
never be assessed against user-agent ... UNLESS then /ALL qualifier is applied. 
When /ALL is used comparison continues against all filter strings even if one
or more do not match.  This is useful for displaying the totals that match each
of multiple filter critera (not much use against a single one).

Although part of the WASD package there is no reason why this shouldn't be of
use with other packages if desired.


EXAMPLES
--------
Requires a foreign verb or such (shorter version might save keystrokes :^)

  $ QDLOG == "$HT_EXE:QDLOGSTATS"
  $ QDLOGSTATS == "$HT_EXE:QDLOGSTATS"

As the character sequences '/' then '*', and '*' then '/' are reserved in C,
the following examples have '@' substituted for '*'.  Please read '*' for all
instances of '@'!

1)  Records from September 1999.

    $ QDLOGSTATS HT_LOGS:@1999@.LOG /DATE="@/SEP/1999@"

2)  Records where the browser was an X-based Netscape Navigator

    $ QDLOGSTATS HT_LOGS:@.LOG /USERAGENT=@MOZILLA@X11@

3)  Records of POST method requests

    $ QDLOGSTATS HT_LOGS:@.LOG /METHOD=POST

4)  Records requesting a particular path

    $ QDLOGSTATS HT_LOGS:@.LOG /PATH="/cgi-bin/@"

5)  Records with a query string (second excludes those containing "grotty")

    $ QDLOGSTATS HT_LOGS:@.LOG /QUERY="%@"
    $ QDLOGSTATS HT_LOGS:@.LOG /QUERY="+{%@}-{@grotty@}"

6)  Records requesting a particular path, excluding the help script

    $ QDLOGSTATS HT_LOGS:@.LOG /PATH="+{/cgi-bin/@}-{/cgi-bin/help/@}"

7)  Select proxy records requesting (a) particular site(s)

    $ QDLOGSTATS HT_LOGS:@8080@.LOG /PATH="http://@.compaq.com@"
    $ QDLOGSTATS HT_LOGS:@8080@.LOG /METHOD=POST /PATH="http://@sex@.@/@"

8)  Records where the request was authenticated

    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="-{-}"
    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="DANIEL"

9)  Records where the request was authenticated, excluding DANIEL

    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="-{-}-{DANIEL}"


QUALIFIERS
----------
/ALL                  compare all records to all filters
/DBUG                 turns on all "if (Debug)" statements
/AUTHUSER=            pattern (any authenticated username)
/CLIENT=              pattern (client host name or IP address)
/DATETIME=            pattern ("11/Jun/1999:14:08:49 +0930")
/METHOD=              HTTP method ("GET", "POST", "HEAD", etc.)
/OUTPUT=              file specification
/PATH=                pattern (URL path component only)
/PROGRESS[=integer]   show progress during processing
/QUERY=               pattern (URL query component only)
/REFERER=             pattern (HTTP "Referer:" field, COMBINED only)
/USERAGENT=           pattern (HTTP "User-Agent:", COMBINED only)
/VIEW[=type]          display matching records, where <type> can be
                      ALL, MATCH (default) or NOMATCH


BUILD DETAILS
-------------
See BUILD_QDLOGSTATS.COM procedure.


COPYRIGHT
---------
Copyright (c) 2000 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update SoftwareID as well!)
---------------
14-NOV-2000  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "QDLOGSTATS AXP-1.0.0";
#else
   char SoftwareID [] = "QDLOGSTATS VAX-1.0.0";
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unixio.h>
#include <unixlib.h>

/* VMS-related header files */
#include <descrip.h>
#include <rms.h>
#include <lib$routines.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>
#include <starlet.h>

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
/* used in modulus for every so many records progress */
#define PROGRESS_RECORDS 1000
/* wrap every so many characters */
#define PROGRESS_WRAP 80

/* maximum log line/record length */
#define MAX_LINE_LENGTH 2048

#define DEFAULT_LOG_FILESPEC "HT_LOGS:*ACCESS*.LOG;0"

#define VIEW_NONE    0
#define VIEW_ALL     1
#define VIEW_MATCH   2
#define VIEW_NOMATCH 3

char  Utility [] = "QDLOGSTATS";

boolean  Debug,
         FilterOnAll;

int  AuthUserFilterAcceptCount,
     AuthUserFilterRejectCount,
     ClientFilterAcceptCount,
     ClientFilterRejectCount,
     CombinedLogRecordCount,
     CommonLogRecordCount,
     DateTimeFilterRejectCount,
     DateTimeFilterAcceptCount,
     FileCountTotal,
     MethodFilterAcceptCount,
     MethodFilterRejectCount,
     PathFilterAcceptCount,
     PathFilterRejectCount,
     ProgressCount,
     QueryFilterAcceptCount,
     QueryFilterRejectCount,
     RecordCount,
     RecordCountTotal,
     SuspectRecordCountTotal,
     RequestCountTotal,
     RefererFilterAcceptCount,
     RefererFilterRejectCount,
     ShowProgress,
     UserAgentFilterAcceptCount,
     UserAgentFilterRejectCount,
     ViewRecords;

int  StatusCodeCount [6];

float  ByteCountTotal;

char  *AuthUserFilterPtr,
      *ClientFilterPtr,
      *DateTimeFilterPtr,
      *MethodFilterPtr,
      *OutputPtr,
      *PathFilterPtr,
      *QueryFilterPtr,
      *RefererFilterPtr,
      *UserAgentFilterPtr;

char  FileSpec [256];

/* required function prototypes */
boolean  FilterThisOut (char*, char*);
int GetParameters ();
int ProcessLogFile (char*);
int ProcessLogRecord (char*);
int SearchFileSpec (char*);
int ShowHelp ();
char* SysGetMsg (int);
int strsame (char*, char*, int);

/*****************************************************************************/
/*
*/

main ()

{
   char  AuthUserString [64],
         ByteString [64],
         ClientString [64],
         DateTimeString [64],
         MethodString [64],
         PathString [64],
         QueryString [64],
         RefererString [64],
         UserAgentString [64];
   unsigned long  UnixSeconds;

   /*********/
   /* begin */
   /*********/

   GetParameters ();

   if (OutputPtr != NULL)
      if ((stdout = freopen (OutputPtr, "w", stdout)) == NULL)
         exit (vaxc$errno);

   if (!FileSpec[0])
   {
      ShowHelp ();
      exit (SS$_NORMAL);
   }

   time (&UnixSeconds);
   fprintf (stdout, "%s, %s", SoftwareID, ctime (&UnixSeconds));

   SearchFileSpec (FileSpec);

   if (ByteCountTotal > 1000000000.0)
   {
      ByteCountTotal /= 1000000000.0;
      sprintf (ByteString, "GBytes ....... %.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000000.0)
   {
      ByteCountTotal /= 1000000.0;
      sprintf (ByteString, "MBytes ....... %.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000.0)
   {
      ByteCountTotal /= 1000.0;
      sprintf (ByteString, "KBytes ....... %.3f", ByteCountTotal);
   }
   else
      sprintf (ByteString, "Bytes ........ %.0f", ByteCountTotal);

   if (ShowProgress && FileCountTotal) fputs ("\n", stdout);

   if (FilterOnAll)
   {
      if (ClientFilterPtr == NULL)
         strcpy (ClientString, "<none>");
      else
         sprintf (ClientString, "  (%d nomatch, %d match)",
                  ClientFilterRejectCount, ClientFilterAcceptCount);
      if (AuthUserFilterPtr == NULL)
         strcpy (AuthUserString, "<none>");
      else
         sprintf (AuthUserString, "  (%d nomatch, %d match)",
                  AuthUserFilterRejectCount, AuthUserFilterAcceptCount);
      if (DateTimeFilterPtr == NULL)
         strcpy (DateTimeString, "<none>");
      else
         sprintf (DateTimeString, "  (%d nomatch, %d match)",
                  DateTimeFilterRejectCount, DateTimeFilterAcceptCount);
      if (MethodFilterPtr == NULL)
         strcpy (MethodString, "<none>");
      else
         sprintf (MethodString, "  (%d nomatch, %d match)",
                  MethodFilterRejectCount, MethodFilterAcceptCount);
      if (PathFilterPtr == NULL)
         strcpy (PathString, "<none>");
      else
         sprintf (PathString, "  (%d nomatch, %d match)",
                  PathFilterRejectCount, PathFilterAcceptCount);
      if (QueryFilterPtr == NULL)
         strcpy (QueryString, "<none>");
      else
         sprintf (QueryString, "  (%d nomatch, %d match)",
                  QueryFilterRejectCount, QueryFilterAcceptCount);
      if (RefererFilterPtr == NULL)
         strcpy (RefererString, "<none>");
      else
         sprintf (RefererString, "  (%d nomatch, %d match)",
                  RefererFilterRejectCount, RefererFilterAcceptCount);
      if (UserAgentFilterPtr == NULL)
         strcpy (UserAgentString, "<none>");
      else
         sprintf (UserAgentString, "  (%d nomatch, %d match)",
                  UserAgentFilterRejectCount, UserAgentFilterAcceptCount);
   }
   else
   {
      if (ClientFilterPtr == NULL)
         strcpy (ClientString, "<none>");
      else
         sprintf (ClientString, "  (%d nomatch)", ClientFilterRejectCount);
      if (AuthUserFilterPtr == NULL)
         strcpy (AuthUserString, "<none>");
      else
         sprintf (AuthUserString, "  (%d nomatch)", AuthUserFilterRejectCount);
      if (DateTimeFilterPtr == NULL)
         strcpy (DateTimeString, "<none>");
      else
         sprintf (DateTimeString, "  (%d nomatch)", DateTimeFilterRejectCount);
      if (MethodFilterPtr == NULL)
         strcpy (MethodString, "<none>");
      else
         sprintf (MethodString, "  (%d nomatch)", MethodFilterRejectCount);
      if (PathFilterPtr == NULL)
         strcpy (PathString, "<none>");
      else
         sprintf (PathString, "  (%d nomatch)", PathFilterRejectCount);
      if (QueryFilterPtr == NULL)
         strcpy (QueryString, "<none>");
      else
         sprintf (QueryString, "  (%d nomatch)", QueryFilterRejectCount);
      if (RefererFilterPtr == NULL)
         strcpy (RefererString, "<none>");
      else
         sprintf (RefererString, "  (%d nomatch)", RefererFilterRejectCount);
      if (UserAgentFilterPtr == NULL)
         strcpy (UserAgentString, "<none>");
      else
         sprintf (UserAgentString, "  (%d nomatch)", UserAgentFilterRejectCount);
   }

   fprintf (stdout,
"\n\
Log .......... %s  (%d processed)\n\
Client ....... %s%s%s%s\n\
Auth User .... %s%s%s%s\n\
Date/Time .... %s%s%s%s\n\
Method ....... %s%s%s%s\n\
Path ......... %s%s%s%s\n\
Query ........ %s%s%s%s\n\
Referer ...... %s%s%s%s\n\
User Agent ... %s%s%s%s\n\
\n\
Records ...... %d  (%d suspect, %d common, %d combined)\n\
Requests ..... %d\n\
Responses .... 1nn:%d  2nn:%d  3nn:%d  4nn:%d  5nn:%d  ?:%d\n\
%s\n",
            FileSpec, FileCountTotal,
            ClientFilterPtr == NULL ? "" : "\"" ,
            ClientFilterPtr == NULL ? "" : ClientFilterPtr,
            ClientFilterPtr == NULL ? "" : "\"" ,
            ClientString,
            AuthUserFilterPtr == NULL ? "" : "\"" ,
            AuthUserFilterPtr == NULL ? "" : AuthUserFilterPtr,
            AuthUserFilterPtr == NULL ? "" : "\"" ,
            AuthUserString,
            DateTimeFilterPtr == NULL ? "" : "\"" ,
            DateTimeFilterPtr == NULL ? "" : DateTimeFilterPtr,
            DateTimeFilterPtr == NULL ? "" : "\"" ,
            DateTimeString,
            MethodFilterPtr == NULL ? "" : "\"" ,
            MethodFilterPtr == NULL ? "" : MethodFilterPtr,
            MethodFilterPtr == NULL ? "" : "\"" ,
            MethodString,
            PathFilterPtr == NULL ? "" : "\"" ,
            PathFilterPtr == NULL ? "" : PathFilterPtr,
            PathFilterPtr == NULL ? "" : "\"" ,
            PathString,
            QueryFilterPtr == NULL ? "" : "\"" ,
            QueryFilterPtr == NULL ? "" : QueryFilterPtr,
            QueryFilterPtr == NULL ? "" : "\"" ,
            QueryString,
            RefererFilterPtr == NULL ? "" : "\"" ,
            RefererFilterPtr == NULL ? "" : RefererFilterPtr,
            RefererFilterPtr == NULL ? "" : "\"" ,
            RefererString,
            UserAgentFilterPtr == NULL ? "" : "\"" ,
            UserAgentFilterPtr == NULL ? "" : UserAgentFilterPtr,
            UserAgentFilterPtr == NULL ? "" : "\"" ,
            UserAgentString,
            RecordCountTotal, SuspectRecordCountTotal,
            CommonLogRecordCount, CombinedLogRecordCount,
            RequestCountTotal,
            StatusCodeCount[1],
            StatusCodeCount[2],
            StatusCodeCount[3],
            StatusCodeCount[4],
            StatusCodeCount[5],
            StatusCodeCount[0],
            ByteString);

   exit (SS$_NORMAL);
}

/****************************************************************************/
/*
Search against the command line file specification, passing each file found to
be processed.
*/ 

SearchFileSpec (char *FileSpec)

{
   int  status,
        FileNameLength,
        Length;
   char  FileName [256],
         ExpFileName [256];
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchFileSpec() |%s|\n", FileSpec);

   /* initialize the file access block */
   SearchFab = cc$rms_fab;
   SearchFab.fab$l_dna = DEFAULT_LOG_FILESPEC;
   SearchFab.fab$b_dns = sizeof(DEFAULT_LOG_FILESPEC)-1;
   SearchFab.fab$l_fna = FileSpec;
   SearchFab.fab$b_fns = strlen(FileSpec);
   SearchFab.fab$l_nam = &SearchNam;

   SearchNam = cc$rms_nam;
   SearchNam.nam$l_esa = ExpFileName;
   SearchNam.nam$b_ess = sizeof(ExpFileName)-1;
   SearchNam.nam$l_rsa = FileName;
   SearchNam.nam$b_rss = sizeof(FileName)-1;

   if (VMSnok (status = sys$parse (&SearchFab, 0, 0)))
      exit (status);

   while (VMSok (status = sys$search (&SearchFab, 0, 0)))
   {
      *SearchNam.nam$l_ver = '\0';
      if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

      ProcessLogFile (FileName);

      FileCountTotal++;
      *SearchNam.nam$l_ver = ';';
   }

   if (VMSnok (status) && status != RMS$_NMF) exit (status);
}

/****************************************************************************/
/*
Open the log file, read and process each record from it, close the file!
*/ 

int ProcessLogFile (char *FileName)

{
   boolean  accepted;
   int  status;
   char  Line [MAX_LINE_LENGTH],
         LineBuffer [MAX_LINE_LENGTH];
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessLogFile() |%s|\n", FileName);

   if (ShowProgress && !ViewRecords)
   {
      if (ProgressCount == PROGRESS_WRAP)
      {
         ProgressCount = 0;
         fputs ("\n", stdout);
      }
      ProgressCount++;
      fputs ("+", stdout);
   }

   RecordCount = 0;

   FilePtr = fopen (FileName, "r", "shr=get");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return (status);
   }

   while (fgets (Line, sizeof(Line), FilePtr) != NULL)
   {
      RecordCount++;
      RecordCountTotal++;

      if (ViewRecords == VIEW_NONE)
      {
         if (ShowProgress && !(RecordCount % ShowProgress))
         {
            if (ProgressCount == PROGRESS_WRAP)
            {
               ProgressCount = 0;
               fputs ("\n", stdout);
            }
            ProgressCount++;
            fputs (".", stdout);
         }
      }
      else
      if (ViewRecords == VIEW_ALL)
         fputs (Line, stdout);
      else
         strcpy (LineBuffer, Line);

      if (Debug) fprintf (stdout, "|%s|\n", Line);
      accepted = ProcessLogRecord (Line);

      if (ViewRecords == VIEW_MATCH && accepted)
         fputs (LineBuffer, stdout);
      else
      if (ViewRecords == VIEW_NOMATCH && !accepted)
         fputs (LineBuffer, stdout);
   }

   fclose (FilePtr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Process a single record (line) from the log file.
Common: 'client rfc891 authuser date/time request status bytes'
Combined: 'client rfc891 authuser date/time request status bytes referer agent'
*/

boolean ProcessLogRecord (char *Line)

{
   boolean  RejectRecord;
   char  *cptr, *sptr,
         *AuthUserPtr,
         *ClientPtr,
         *BytesPtr,
         *DateTimePtr,
         *RefererPtr,
         *RequestPtr,
         *StatusPtr,
         *UserAgentPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessLogRecord()\n");

   cptr = Line;

   /* client */
   ClientPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* remote ident (RFC891) */
   while (*cptr && isspace(*cptr)) cptr++;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* authorized user */
   while (*cptr && isspace(*cptr)) cptr++;
   AuthUserPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* date/time */
   while (*cptr && *cptr != '[') cptr++;
   if (*cptr) cptr++;
   DateTimePtr = cptr;
   while (*cptr && *cptr != ']') cptr++;
   if (*cptr) *cptr++ = '\0';

   /* request ("method path?query") */
   while (*cptr && *cptr != '\"') cptr++;
   if (*cptr) cptr++;
   RequestPtr = cptr;
   while (*cptr && *cptr != '\"') cptr++;
   if (*cptr) *cptr++ = '\0';

   /* HTTP response status */
   while (*cptr && isspace(*cptr)) cptr++;
   StatusPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* bytes transmitted */
   while (*cptr && isspace(*cptr)) cptr++;
   BytesPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* referer (only for COMBINED) */
   RefererPtr = NULL;
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr == '\"')
   {
      cptr++;
      RefererPtr = cptr;
      while (*cptr && *cptr != '\"') cptr++;
   }
   else
   if (*cptr)
   {
      RefererPtr = cptr;
      while (*cptr && !isspace(*cptr)) cptr++;
   }
   if (*cptr) *cptr++ = '\0';

   /* user agent (only for COMBINED) */
   UserAgentPtr = NULL;
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr == '\"')
   {
      cptr++;
      UserAgentPtr = cptr;
      while (*cptr && *cptr != '\"') cptr++;
   }
   else
   if (*cptr)
   {
      UserAgentPtr = cptr;
      while (*cptr && !isspace(*cptr)) cptr++;
   }
   if (*cptr) *cptr++ = '\0';

   if (!ClientPtr[0] ||
       !DateTimePtr[0] ||
       !RequestPtr[0] ||
       !StatusPtr[0] ||
       !BytesPtr)
   {
      SuspectRecordCountTotal++;
      return (false);
   }

   if (RefererPtr == NULL && UserAgentPtr == NULL)
      CommonLogRecordCount++;
   else
      CombinedLogRecordCount++;

   RejectRecord = false;

   if (AuthUserFilterPtr != NULL)
   {
      if (FilterThisOut (AuthUserPtr, AuthUserFilterPtr))
      {
         AuthUserFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         AuthUserFilterAcceptCount++;
   }
   if (ClientFilterPtr != NULL)
   {
      if (FilterThisOut (ClientPtr, ClientFilterPtr))
      {
         ClientFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         ClientFilterAcceptCount++;
   }
   if (DateTimeFilterPtr != NULL)
   {
      if (FilterThisOut (DateTimePtr, DateTimeFilterPtr))
      {
         DateTimeFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         DateTimeFilterAcceptCount++;
   }
   if (MethodFilterPtr != NULL)
   {
      for (cptr = RequestPtr; *cptr && *cptr != ' '; cptr++);
      /* terminate path at query string (should be necessary) */
      if (!*cptr) cptr == NULL;
      if (cptr != NULL) *cptr = '\0';
      if (FilterThisOut (RequestPtr, MethodFilterPtr))
      {
         MethodFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         MethodFilterAcceptCount++;
      if (cptr != NULL) *cptr = ' ';
   }
   if (PathFilterPtr != NULL)
   {
      /* skip over the method */
      for (cptr = RequestPtr; *cptr && *cptr != ' '; cptr++);
      while (*cptr && *cptr == ' ') cptr++;
      /* find the start of any query string */
      for (sptr = cptr; *sptr && *sptr != '?'; sptr++);
      /* terminate path at query string if necessary */
      if (!*sptr) sptr == NULL;
      if (sptr != NULL) *sptr = '\0';
      if (FilterThisOut (cptr, PathFilterPtr))
      {
         PathFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         PathFilterAcceptCount++;
      /* return request to pristine condition :^) */
      if (sptr != NULL) *sptr = '?';
   }
   if (QueryFilterPtr != NULL)
   {
      /* find the start of any query string */
      for (cptr = RequestPtr; *cptr && *cptr != '?'; cptr++);
      if (*cptr) cptr++;
      if (FilterThisOut (cptr, QueryFilterPtr))
      {
         QueryFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         QueryFilterAcceptCount++;
   }
   if (RefererFilterPtr != NULL)
   {
      /* if the record did not have a COMBINED referer field then reject */
      if (RefererPtr == NULL ||
          FilterThisOut (RefererPtr, RefererFilterPtr))
      {
         RefererFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         RefererFilterAcceptCount++;
   }
   if (UserAgentFilterPtr != NULL)
   {
      /* if the record did not have a COMBINED user agent field then reject */
      if (UserAgentPtr == NULL ||
          FilterThisOut (UserAgentPtr, UserAgentFilterPtr))
      {
         UserAgentFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         UserAgentFilterAcceptCount++;
   }

   if (RejectRecord) return (false);

   RequestCountTotal++;
   ByteCountTotal += (float)atoi(BytesPtr);
   switch (StatusPtr[0])
   {
      case '1' : StatusCodeCount[1]++; break;
      case '2' : StatusCodeCount[2]++; break;
      case '3' : StatusCodeCount[3]++; break;
      case '4' : StatusCodeCount[4]++; break;
      case '5' : StatusCodeCount[5]++; break;
      default  : StatusCodeCount[0]++;
   }

   return (true);
}

/*****************************************************************************/
/*
Using decc$match_wild() match the pattern against the string.  Parses out the
"+{}" and "-{}" kludges from composite expressions and does the appropriate
boolean processing to return a result.
*/

boolean FilterThisOut
(
char *StringPtr,
char *PatternPtr
)
{
   boolean  Matched,
            NegateMatch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FilterThisOut()\n");

   for (cptr = StringPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   cptr = PatternPtr;
   while (cptr != NULL && *cptr)
   {
      if (*(unsigned short*)cptr == '+{')
      {
         NegateMatch = false;
         cptr += 2;
      }
      else
      if (*(unsigned short*)cptr == '-{')
      {
         NegateMatch = true;
         cptr += 2;
      }
      else
         return (!decc$match_wild (StringPtr, PatternPtr));

      sptr = cptr;
      while (*cptr && *cptr != '}') cptr++;
      if (!*cptr) cptr = NULL;
      if (cptr != NULL) *cptr = '\0';
      Matched = decc$match_wild (StringPtr, sptr);
      if (cptr != NULL) *cptr++ = '}';
      if (NegateMatch) Matched = !Matched;
      if (!Matched) return (true);
   }

   return (false);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   int  status;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("QDLOG$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (strsame (aptr, "/ALL", -1))
      {
         FilterOnAll = true;
         continue;
      }
      if (strsame (aptr, "/AUTHUSER=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (AuthUserFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/CLIENT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (ClientFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/DATETIME=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (DateTimeFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/METHOD=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (MethodFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/OUTPUT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         OutputPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/PATH=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (PathFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/PROGRESS=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         ShowProgress = atoi(cptr);
         if (ShowProgress <= 0) ShowProgress = PROGRESS_RECORDS;
         continue;
      }
      if (strsame (aptr, "/QUERY=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (QueryFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/REFERER=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (RefererFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/USERAGENT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (UserAgentFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/VIEW=", 4))
      {
         ViewRecords = VIEW_MATCH;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         if (strsame (cptr, "ALL", 3))
            ViewRecords = VIEW_ALL;
         else
         if (strsame (cptr, "MATCH", 3))
            ViewRecords = VIEW_MATCH;
         else
         if (strsame (cptr, "NOMATCH", 3))
            ViewRecords = VIEW_NOMATCH;
         else
         {
            fprintf (stdout, "%%%s-E-INVPARM, invalid parameter\n \\%s\\\n",
                     Utility, cptr);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      if (!FileSpec[0])
      {
         sptr = FileSpec;
         for (cptr = aptr; *cptr; *sptr++ = toupper(*cptr++));
         *sptr = '\0';
         continue;
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
*/
 
char* SysGetMsg (int StatusValue)
 
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   short int  Length;
 
   sys$getmsg (StatusValue, &Length, &MessageDsc, 0, 0);
   Message[Length] = '\0';
   if (Debug) fprintf (stdout, "SysGetMsg() |%s|\n", Message);
   return (Message);
}
 
/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
*/

int ShowHelp ()

{
   fprintf (stdout,
"Usage for Quick and Dirty LOG STATisticS (%s)\n\
\n\
$ QDLOGSTATS <log-file-spec> [<filters>] [<other qualifiers>]\n\
\n\
Utility to extract very elementary statistics from Web server common/combined\n\
format log files.  A number of filters allow subsets of the log contents to be\n\
selected using simple wildcard expressions.  Strings are NOT case-sensitive.\n\
Optionally, log file records can be viewed as processed, or a simple progress\n\
indicator can be displayed (\"+\" for each file, \".\" per 1000 records thereof).\n\
\n\
/ALL /AUTHUSER=filter /CLIENT=filter /DATETIME=filter /METHOD=filter\n\
/OUTPUT=file /PATH=filter /PROGRESS[=integer] /QUERY=filter /RFERER=filter\n\
/USERAGENT=filter /VIEW[=MATCH(D)|ALL|NOMATCH]\n\
\n\
Usage examples:\n\
\n\
$ QDLOGSTATS == \"$dir:QDLOGSTATS\"\n\
$ QDLOGSTATS HT_LOGS:*.LOG /PROGRESS\n\
$ QDLOGSTATS HT_LOGS:*NOV*ACCESS* /PATH=\"/CGI-BIN/*\" /QUERY=\"-{-}\"\n\
$ QDLOGSTATS HT_LOGS:*ACCESS*.LOG /METHOD=POST /DATE=\"*/NOV/2000/*\"\n\
$ QDLOGSTATS HT_LOGS:* /REFERER=*COMPAQ.COM*\n\
$ QDLOGSTATS HT_LOGS:*.LOG /USERAGENT=*MOZILLA*X11*\n\
\n",
   SoftwareID);

   return (SS$_NORMAL);
}

/*****************************************************************************/

